/******************************************************************************

                               Copyright (c) 2010
                            Lantiq Deutschland GmbH
                     Am Campeon 3; 85579 Neubiberg, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include "pmd.h"


/**
   \defgroup LQ_PMD Power Management Daemon
   @{
*/

/**
   \defgroup LQ_PMD_PRG Program Guide
   <center>
   <B>"Power Management Daemon ( PMD )" </B>.
   </center>    
   \section Purpose Purpose
   The Power Management Daemon(PMD) supervise the current cpu load of the system and decrease 
   the clock frequency of the system to reduce current consumption, if the cpu load signalize 
   a high IDLE value. On the other hand if the PMD detect a high cpu load the clock frequency
   will be increased to ensure a proper behavior of the system.
   
   \section Introduction Introduction
   The PMD is started during bootup as a background process via the startup script <I>/etc/init.d/pmd.sh</I>.\n
   The binary for the daemon is located under <I>/opt/lantiq/bin/pmd</I>.\n
   The daemon reads in a given interval the <I>/proc/stat</I> file and calculate based on this values
   a peak and average cpu load. Dependent on the given thresholds the clock frequency for the
   CPU system and the DDR will be decreased if the cpu load signalize a high idle time. 
   In the other case, if the cpu load signalize a high cpu load, the clock frequency will be increased.\n
   \n
   Supported Power States:
        <table>
        <tr><td>D0</td> <td>on State</td> </tr>
        <tr><td>D1</td> <td>standby</td> </tr>
        <tr><td>D2</td> <td>low power standby</td> </tr>
        <tr><td>D3</td> <td>off State</td> </tr>
     </table>


   The following behavior is supported and implemented:
   -  If the cpu load is below a minimum threshold <B>(CPU_LOAD_AVG_MIN</B>, the clock frequency is 
      reduced in single steps. Dependent on the current power state the next lower power state is requested.
      <B>( D0->D1->D2->D3 )</B>     
   -  If the cpu load is above a maximum threshold <B>(CPU_LOAD_AVG_MAX)</B>, the clock frequency is increased.
      In this case we change always to the <B>onState</B> power state <B>D0</B> to ensure a proper performance of the
      system to process the requested tasks.\n
      Both of these actions are based on an average calculation over a given time spawn.
   -  The third threshold detect a peak load <B>(CPU_LOAD_MAX)</B> of the cpu and immediately increase the 
      clock frequency, without waiting for the average interval.
   
   The average cycles for increasing and decreasing are separated and can have different values.
   Means that we may define a faster response for the clock frequency increase and/or give us more
   time to slow down the system, because this is not time critical.

   \section Usage Usage
   The PMD may called without any parameters. In this case the PMD will use the default values for
   the different threshold values and cycle/average times.\n
   <B>Example:</B> :_><I>pmd</I>\n
   
   A help menue may displayed by calling the pmd with the option -h\n
   <B>Example:</B> :_><I>pmd -h</I>\n
   
   If you want to call the PMD with parameter values different to the default values the call looks like:\n
   <B>Example:</B> :_><I>pmd 20 60 90 200 20 30 1</I>\n
   - 1<SUP>st</SUP> parameter : CPU_LOAD_AVG_MIN
   - 2<SUP>sd</SUP> parameter : CPU_LOAD_AVG_MAX
   - 3<SUP>rd</SUP> parameter : CPU_LOAD_MAX
   - 4<SUP>th</SUP> parameter : CYCLE_TIME
   - 5<SUP>th</SUP> parameter : NUM_CYCLES_UP
   - 6<SUP>th</SUP> parameter : NUM_CYCLES_DOWN
   - 7<SUP>th</SUP> parameter : LOG_LEVEL\n\n
   
   
*/


/**
   \defgroup LQ_PMD_DEBUG Debug utilities
   PMD printf specific's; colored print_message support
   @{
*/
/******************************************************************************/
#define BLINK   "\033[31;1m"
#define RED     "\033[31;1m"
#define YELLOW  "\033[33;1m"
#define GREEN   "\033[32;2m"
#define BLUE    "\033[34;1m"
#define CYAN    "\033[36;2m"
#define DIM     "\033[37;1m"
#define NORMAL  "\033[0m"

#define PMD_ERR(module,fmt,arg...)\
        if(log_level_g > 0){\
        printf(RED #module NORMAL ": " fmt,##arg);\
        }
#define PMD_INFO(module,fmt,arg...)\
        if(log_level_g > 1){\
        printf(GREEN #module NORMAL ": " fmt,##arg);\
        }
#define PMD_DEBUG(module,fmt,arg...)\
        if(log_level_g > 2){\
        printf(BLUE #module NORMAL ": " fmt,##arg);\
        }
/*@}*/
/******************************************************************************/

#define FALSE	(0)                  /**<  definition of return value FALSE */
#define TRUE	(!FALSE)             /**<  definition of return value TRUE  */
#define COC_PMD_VERSION  "1.0.3.0"  /**<  Version code for PMD  */

/* Thresholds defaults */
#define CPU_LOAD_AVG_MIN    20   /**< Threshold default: 20% cpu load average for decreasing clk freq */
#define CPU_LOAD_AVG_MAX    60   /**< Threshold default: 60% cpu load average for increasing clk freq */
#define CPU_LOAD_MAX        90   /**< Threshold default: 90% cpu load current peak for immediately increasing clk freq */

/* Cycle time to calculate the current cpu load */
#define CYCLE_TIME 1000 * 100    /**< Cycle time to calculate the current cpu load. Default cycle time 100msec */

/* Time interval to calculate the average cpu load */
#define NUM_CYCLES_UP      10    /**< average cycle number for increasing clock frequency */
#define NUM_CYCLES_DOWN    20    /**< average cycle number for decreasing clock frequency */


/*=============================================================================*/
/* MODULE GLOBAL VARIABLES                                                     */
/*=============================================================================*/
/** Power state translation */
static char* pmcu_ps[] = {
   "INVALID" ,
   "D0" ,
   "D1" ,
   "D2" ,
   "D3" ,
};
int cpu_load_avg_min_g  = CPU_LOAD_AVG_MIN;  /**< Threshold variable: 20% cpu load average for decreasing clk freq */
int cpu_load_avg_max_g  = CPU_LOAD_AVG_MAX;  /**< Threshold variable: 60% cpu load average for increasing clk freq */
int cpu_load_max_g      = CPU_LOAD_MAX;      /**< Threshold variable: 90% cpu load current peak for immediately increasing clk freq */
int cycle_time_g        = CYCLE_TIME;        /**< Cycle time variable to calculate the current cpu load. Default cycle time 100msec */
int num_cycles_up_g     = NUM_CYCLES_UP;     /**< Cycle counter to calculate the average cpu load for increasing clk freq */
int num_cycles_down_g   = NUM_CYCLES_DOWN;   /**< Cycle counter to calculate the average cpu load for decreasing clk freq */
int log_level_g         = 1;                 /**< Loglevel control for the module: 0=quiet, 1=only ERRORS, 2=+INFO, 3=+DEBUG */

/*=============================================================================*/
/* FUNCTION DEFINITIONS                                                        */
/*=============================================================================*/
/**
   PMD help menu.\n
   Shows an overview about the parameters and the options which are
   supported by the PMD

   \return None
*/
static void print_help(void){
   fprintf (stderr, "Power Management Daemon Version = %s\n",COC_PMD_VERSION);
   fprintf (stderr, "Usage: pmd [OPTIONS]\n\n");
   fprintf (stderr, "OPTIONS:\n");
   fprintf (stderr, "         none  -> pmd will use the default values for\n");
   fprintf (stderr, "                  cycle/average time and thresholds\n");
   fprintf (stderr, "         -h    -> print this help menu\n");
   fprintf (stderr, "         if you want to override any default value\n");
   fprintf (stderr, "         the complete parameter set must be specified in\n");
   fprintf (stderr, "         the given sequence:\n");
   fprintf (stderr, "         CPU_LOAD_AVG_MIN -> threshold for cpu load in %%\n");
   fprintf (stderr, "                             If cpu load is below this value the\n");
   fprintf (stderr, "                             clock frequency will be decreased.\n");
   fprintf (stderr, "                             Default: 20%%\n");
   fprintf (stderr, "         CPU_LOAD_AVG_MAX -> threshold for cpu load in %%\n");
   fprintf (stderr, "                             If cpu load is above this value the\n");
   fprintf (stderr, "                             clock frequency will be increased.\n");
   fprintf (stderr, "                             Default: 60%%\n");
   fprintf (stderr, "         CPU_LOAD_MAX     -> threshold for cpu load in %%\n");
   fprintf (stderr, "                             This value defines the peak cpu load. If \n");
   fprintf (stderr, "                             the cpu load is above this value the \n");
   fprintf (stderr, "                             clock frequency will be immediately increased \n");
   fprintf (stderr, "                             without waiting until the end of the \n");
   fprintf (stderr, "                             average time spawn is reached.\n");
   fprintf (stderr, "                             Default: 90%%\n");
   fprintf (stderr, "         CYCLE_TIME       -> polling cycle in msec. Defines the time  \n");
   fprintf (stderr, "                             how often the /proc/stat file is processed.\n");
   fprintf (stderr, "                             Default: 100\n");
   fprintf (stderr, "         NUM_CYCLES_UP    -> Defines the number of CYCLE_TIME which must be \n");
   fprintf (stderr, "                             passed for average_up calculation. \n");
   fprintf (stderr, "                             'up' means increasing clock freq \n");
   fprintf (stderr, "                             Default: 10\n\n");
   fprintf (stderr, "         NUM_CYCLES_DOWN  -> Defines the number of CYCLE_TIME which must be \n");
   fprintf (stderr, "                             passed for average_down calculation. \n");
   fprintf (stderr, "                             'down' means decreasing clock freq \n");
   fprintf (stderr, "                             Default: 20\n\n");
   fprintf (stderr, "         LOG_LEVEL        -> controls the message level \n");
   fprintf (stderr, "                             0: QUIET, 1:only ERRORS, 2:+INFO, 3:+DEBUG \n");
   fprintf (stderr, "                             Default: 1\n\n");
   fprintf (stderr, "Example using default values: /opt/lantiq/bin/pmd\n");
   fprintf (stderr, "Example calling help        : /opt/lantiq/bin/pmd -h\n");
   fprintf (stderr, "Example overwrite defaults  : /opt/lantiq/bin/pmd 20 60 90 200 20 30 1\n");
}


/**
   The PMD_NextLowerState function return the next lower Powerstate. 
   The steps are defined by the power state definitions: 
   - IFX_PMCU_STATE_D0 -> highest frequency 
   - IFX_PMCU_STATE_D1 -> one step lower 
   - IFX_PMCU_STATE_D2 -> one step lower 
   - IFX_PMCU_STATE_D3 -> lowest frequency 

   \return None
*/
static IFX_PMCU_STATE_t PMD_NextLowerState(IFX_PMCU_STATE_t state)
{
   //printf("dec_clk_freq is called\n");
   if(state == IFX_PMCU_STATE_D3){
      return IFX_PMCU_STATE_INVALID; /* nothing to do; already in D3 state */
   }
   switch (state){
      case IFX_PMCU_STATE_D0:{
         state = IFX_PMCU_STATE_D1;
         break;
      }
      case IFX_PMCU_STATE_D1:{
         state = IFX_PMCU_STATE_D2;
         break;
      }
      case IFX_PMCU_STATE_D2:{
         state = IFX_PMCU_STATE_D3;
         break;
      }
      default:
         break;
   }
   return state;
}


/**
   The PMD_SetState function request a new power state via the PMCU
   ioctl-interface to change the clock frequency.

   \param [in]  state  requested power state (clock frequency)

   \return None
*/
static void PMD_SetState(IFX_PMCU_STATE_t state)
{
   int ret,fd;
   IFX_PMCU_MODULE_STATE_t param={0,0,0};
   
   fd=open("/dev/ifx_pmcu",O_RDWR);
   if(fd<0) {
      PMD_ERR(PMD_,"cannot open device!\n");
      return;
   }
   PMD_INFO(PMD_,"Clock Frequency Change to %s\n",pmcu_ps[state]);
   param.pmcuModule=IFX_PMCU_MODULE_CPU;
   param.pmcuModuleNr=0;
   param.pmcuState=state;
   while(param.pmcuState != IFX_PMCU_STATE_INVALID) {
      ret=ioctl(fd,IFX_PMCU_STATE_REQ,&param);
      if(ret<0){
         PMD_DEBUG(PMD_,"PowerState %s DENIED\n",pmcu_ps[state]);
         param.pmcuState = PMD_NextLowerState(param.pmcuState);
      }else{
         break;   
      }
   }

   close(fd);
   return;
}

/**
   The PMD_GetState function read the current power state from the 
   clock frequency driver, to decide if a further change of the 
   power state is necessary or not.

   \return  current power state of the clock frequency driver.
            Returned values -> IFX_PMCU_STATE_t
*/
static IFX_PMCU_STATE_t PMD_GetState(void)
{
   IFX_PMCU_MODULE_STATE_t param={0,0,IFX_PMCU_STATE_INVALID};
   int ret,fd;

   fd=open("/dev/ifx_pmcu",O_RDWR);
   if(fd<0) {
      PMD_ERR(PMD_,"cannot open device!\n");
      return -1;
   }
   param.pmcuModule=IFX_PMCU_MODULE_CPU;
   param.pmcuModuleNr=0;
   ret=ioctl(fd,IFX_PMCU_STATE_GET,&param);
   if(ret<0){
      PMD_ERR(PMD_,"ioctl IFX_PMCU_STATE_GET failed!\n");
   }
   close(fd);
   return param.pmcuState;
}


/**
   The PMD_IncClkFreq function set the clock frequency always to
   highest value, to insure each time the best permormance for the
   system.

   \return None
*/
static void PMD_IncClkFreq(void)
{
   //printf("set_state is called\n");
   if(PMD_GetState() == IFX_PMCU_STATE_D0){
      return; /* nothing to do; already in D0 state */
   }
   PMD_SetState(IFX_PMCU_STATE_D0);
}


/**
   The PMD_DecClkFreq function reduce the clock frequency bei one step.
   The steps are defined by the power state definitions:
   - IFX_PMCU_STATE_D0 -> highest frequency 
   - IFX_PMCU_STATE_D1 -> one step lower 
   - IFX_PMCU_STATE_D2 -> one step lower 
   - IFX_PMCU_STATE_D3 -> lowest frequency 
   The absolut frequency values depends on the used platform.

   \return None
*/
static void PMD_DecClkFreq(void)
{
   IFX_PMCU_STATE_t state;

   //printf("dec_clk_freq is called\n");
   if(state == IFX_PMCU_STATE_INVALID) {
      state = PMD_GetState();
   }
   
   if(state == IFX_PMCU_STATE_D3){
      return; /* nothing to do; already in D3 state */
   }
   switch (state){
      case IFX_PMCU_STATE_D0:{
         state = IFX_PMCU_STATE_D1;
         break;
      }
      case IFX_PMCU_STATE_D1:{
         state = IFX_PMCU_STATE_D2;
         break;
      }
      case IFX_PMCU_STATE_D2:{
         state = IFX_PMCU_STATE_D3;
         break;
      }
      default:
         break;
   }
   PMD_SetState(state);
}

/**
   The PMD_ReadValues function read the first line from /proc/stat, which 
   aggregates the numer in all of the other "cpu lines".
   These numbers identify the amount of time the CPU has spent performing
   different kinds of work. Time units are in "Jiffies".
   In "used" the accumulated status of all active process are stored.
   "idle" hold the "twiddling thumbs" time.

   \param [in]  fd   file descriptor to /proc/stat
   \param [out] used cpu load value  
   \param [out] idle cpu idle time  

   \return Returns value as follows:
     - TRUE : if successful
     - FALSE: in case of an error
*/
static int PMD_ReadValues (int fd, long int *used, long int *idle)
{
   int n;
   long int ret;
   char *p;
   static char buf[60];

   read (fd, buf, sizeof(buf));
   if (strncmp (buf, "cpu ", 4)) {
      PMD_ERR(PMD_,"Unexpected first line in /proc/stat:\n%s\n", buf);
      return FALSE;
   }
   lseek (fd, 0, SEEK_SET);
   *used = 0;
   for (p = buf + 4, n = 0;n < 7; n ++) {
      ret = strtol (p, &p, 10);
      //PMD_DEBUG(PMD_,"ret=%ld\n",ret);
      if(n == 3){
         *idle = ret;
      }else{
         *used += ret;
      }
   }
   return TRUE;
}
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

/**
   The main function poll's in an endless loop the /proc/stat file and calculate
   based on this vales a peak and average cpu load. Dependent on give thresholds
   the clock frequency for the CPU system and the DDR will be decreased if the
   cpu load signalize a high idle time. In the other case, if the cpu load signalize
   a high cpu load, the clock frequency will be increased.

   \return None
*/
int main (int argc, char *argv[])
{
   int fd;
   int ret = TRUE;
   long int total[2] = {0,0}, used[3] = {0,0,0}, idle[3] = {0,0,0}; /* [0]=is current value, [1]=previous value, [2]=diff */
   int cpu_load[5] = {0,0,0,0,0}; /* [0]=current cpu load in %, [1]=accumulated_up cpu load in %, [2]=average_up cpu load in % */
                                  /* [3]=accumulated_down cpu load in %, [4]=average_down cpu load in % */
   int avg_cnt_up = 0;   /* average loop counter increase frequency */
   int avg_cnt_down = 0; /* average loop counter decrease frequency */

   if(argc == 8){
      /* overwrite the default values cycle/average time and thresholds */
      cpu_load_avg_min_g  = strtol (argv[1], NULL, 10);
      cpu_load_avg_max_g  = strtol (argv[2], NULL, 10);
      cpu_load_max_g      = strtol (argv[3], NULL, 10);
      cycle_time_g        = (strtol (argv[4], NULL, 10)) * 1000; /* convert to usec */
      num_cycles_up_g     = strtol (argv[5], NULL, 10);
      num_cycles_down_g   = strtol (argv[6], NULL, 10);
      log_level_g         = strtol (argv[7], NULL, 10);
   }else{
      if(argc != 1){
         print_help();
         return TRUE;
      }
   }
   
   fd = open ("/proc/stat", O_RDONLY);
   /* must be read first to initialize the previous values, otherwise
      first compare between current and previous leads to 100% cpu load.*/
   if (! PMD_ReadValues (fd, &used[0], &idle[0])) {
      PMD_ERR(PMD_,"Error during read /proc/stat\n");;
   }else{
      total[1] = idle[0]  + used[0];
      idle[1]  = idle[0];
   }

   for (;;) {
      if (! PMD_ReadValues (fd, &used[0], &idle[0])) {
         ret = FALSE;
         break;	
      }else{
         total[0] = idle[0]  + used[0];
         total[2] = total[0] - total[1];  /* total diff */
         idle[2]  = idle[0]  - idle[1];   /* idle diff */
         /* Remember the total and idle CPU times for the next check */
         total[1] = total[0];
         idle[1]  = idle[0];

         if(total[2] == 0){
            goto wait_cycle; /* avoid division by NULL at the beginning */
         }
         cpu_load[0] = ((1000*(total[2] - idle[2]) / total[2]) + 5) / 10;
         //PMD_DEBUG(PMD_,"CPU load = %d%%\n",cpu_load[0]);
         /* check if current cpu load exceeds the peak value */
         if(cpu_load[0] > cpu_load_max_g){
            PMD_DEBUG(PMD_,"CPU load peak detected;\n");
            /* increase clock frequency */
            PMD_IncClkFreq();
            /* reset avg calculation */
            avg_cnt_up = 0;
            cpu_load[1] = 0;
            goto wait_cycle;
         }
         cpu_load[1] = cpu_load[1] + cpu_load[0];  /* accumulate the cpu load during the average up interval */
         cpu_load[3] = cpu_load[3] + cpu_load[0];  /* accumulate the cpu load during the average down interval */
         avg_cnt_up++;
         avg_cnt_down++;
         if(avg_cnt_up > num_cycles_up_g){
            /* average calculation */
            cpu_load[2] = cpu_load[1] / num_cycles_up_g;
            PMD_DEBUG(PMD_,"CPU_avg_load_up: %d%%\n",cpu_load[2]);
            /* reset avg calculation */
            avg_cnt_up = 0;
            cpu_load[1] = 0;
            if(cpu_load[2] > cpu_load_avg_max_g){
               PMD_DEBUG(PMD_,"CPU load > CPU_LOAD_AVG_MAX\n");
               /* increase clock frequency */
               PMD_IncClkFreq();
            }
         }
         if(avg_cnt_down > num_cycles_down_g){
            cpu_load[4] = cpu_load[3] / num_cycles_down_g;
            PMD_DEBUG(PMD_,"CPU_avg_load_down: %d%%\n",cpu_load[4]);
            /* reset avg calculation */
            avg_cnt_down = 0;
            cpu_load[3] = 0;
            if(cpu_load[4] < cpu_load_avg_min_g){
               PMD_DEBUG(PMD_,"CPU load < CPU_LOAD_AVG_MIN\n");
               /* decrease clock frequency */
               PMD_DecClkFreq();
            }
         }
         wait_cycle:
         usleep (cycle_time_g);
      }
   }
   close(fd);
   return ret;
}
/*@}*/ /* LQ_PMD */
